#include <SPI.h>
#include <LoRa.h>
#include <TimeOut.h>
#include <CommandParser.h>
#include "BluetoothSerial.h"  //Header File for Serial Bluetooth
#include <esp_bt.h>
//Define who will use this device
//#define CC
#define BB

BluetoothSerial ESP_BT;  //Object for Bluetooth
boolean confirmRequestPending = true;

int incoming;

//define the pins used by the transceiver module
#define rst 26
#define dio0 25
#define ant_switch 33     //1 for transmitting, 0 for receiving
#define HSPI_MISO   12
#define HSPI_MOSI   13
#define HSPI_SCLK   14
#define HSPI_SS     27

//Buzzer and Vibration
const int PIN_Buzzer =  16;
const int PIN_Vibration =  17;

//LED R G B
const int PIN_LED_R =15;
const int PIN_LED_G =2;
const int PIN_LED_B =4;

//Button
const int PushButton =36;

//static const int spiClk = 10000000; // 10 MHz
SPIClass * hspi = NULL;

int counter = 0;

//Lora Varialbles
String outgoing;              // outgoing message
byte msgCount = 0;            // count of outgoing messages
#ifdef BB
byte localAddress = 0xBB;     // address of this device
#endif
#ifdef CC
byte localAddress = 0xCC;     // address of this device
#endif
byte destination = 0xFF;      // destination to send to
long lastSendTime = 0;        // last send time
int interval = 2000;          // interval between sends
byte lastActID = 255;         //last ACD ID
enum msgType {msg=0, buzzer=1, vibration=2, buzzer_vibration=3, ack=5} ;
String incomingString ="";

//Lora Switches
bool LoraQualityInfo = false;
byte LocalNotificationConfig = 2; //0 for mute, 1 for buzzer, 2 for vibration, 3 for buzzer+vibration

//Lora ACK Timeout
TimeOut ACKTimeout;
void callbackACKTimeout();
int retryMsg = 0; //0 for timeout, need to resent in xx ms, 1 for resending, 2 for reset to 0

//Local Notification Timeout
TimeOut NTFTimeout;
void callbackNTFTimeout();
#define NTFLocalRingTime 200
#define NTFRemoteRingTime 1000

//Command Parser
typedef CommandParser<5, 3, 10, 255, 255> MyCommandParser;
MyCommandParser parser;



void setup() {
  //initialize Serial Monitor
  Serial.begin(115200);
  while (!Serial);
  Serial.println("LoRa Sender");

  //Init SPI
  hspi = new SPIClass(HSPI);
  hspi->begin(HSPI_SCLK, HSPI_MISO, HSPI_MOSI, HSPI_SS); //SCLK, MISO, MOSI, SS
  pinMode(HSPI_SS, OUTPUT); //HSPI SS
  pinMode(ant_switch, OUTPUT); //RF switch control
  digitalWrite(ant_switch, 0); //Receiving mode as def

  //setup LoRa transceiver module
  LoRa.setPins(HSPI_SS, rst, dio0);
  LoRa.setSPI(*hspi);

  while (!LoRa.begin(915E6)) {
    Serial.println(".");
    delay(500);
  }

  //Supported values are 2 to 20 for PA_OUTPUT_PA_BOOST_PIN, and 0 to 14 for PA_OUTPUT_RFO_PIN.
  LoRa.setTxPower(17,PA_OUTPUT_PA_BOOST_PIN);
  //Supported values are 7.8E3, 10.4E3, 15.6E3, 20.8E3, 31.25E3, 41.7E3, 62.5E3, 125E3, 250E3, and 500E3.
  LoRa.setSignalBandwidth(125E3);
  //Supported values are between 6 and 12. If a spreading factor of 6 is set, implicit header mode must be used to transmit and receive packets.
  LoRa.setSpreadingFactor(12);
  //Supported values are between 5 and 8, these correspond to coding rates of 4/5 and 4/8. The coding rate numerator is fixed at 4.
  LoRa.setCodingRate4(5);
  //preamble length in symbols, defaults to 8
  LoRa.setPreambleLength(8);

  Serial.println("LoRa Initializing OK!");


  //Init BT
  ESP_BT.setTimeout(1000);
  ESP_BT.enableSSP();
  ESP_BT.onConfirmRequest(BTConfirmRequestCallback);
  ESP_BT.onAuthComplete(BTAuthCompleteCallback);
#ifdef BB
  ESP_BT.begin("MSG_BB");  //Name of your Bluetooth Signal
#endif
#ifdef CC
  ESP_BT.begin("MSG_CC");  //Name of your Bluetooth Signal
#endif

  //Setting Bluetooth TX Power
  esp_power_level_t min,max;
  esp_ble_tx_power_set(ESP_BLE_PWR_TYPE_ADV,ESP_PWR_LVL_P9 );
  esp_bredr_tx_power_set(ESP_PWR_LVL_P6,ESP_PWR_LVL_P9); //6dbm to 9dbm
  delay(200);
  esp_bredr_tx_power_get(&min,&max);
  Serial.printf("Bluetooth TX Power: min %d max %d\r\n",min,max);

  Serial.println("Bluetooth Device is Ready to Pair");

  //Init Buzzer and Vibration
  pinMode(PIN_Buzzer, OUTPUT);
  pinMode(PIN_Vibration, OUTPUT);

  //Init Button
  pinMode(PushButton, INPUT);

  //Init LED R G B
  pinMode(PIN_LED_R, OUTPUT);
  pinMode(PIN_LED_G, OUTPUT);
  pinMode(PIN_LED_B, OUTPUT);

  digitalWrite(PIN_LED_R, HIGH);//Red LED OFF
  digitalWrite(PIN_LED_G, HIGH);//Gree LED OFF
  digitalWrite(PIN_LED_B, HIGH);//Blue LED OFF

  //register commands
  parser.registerCommand("ATS+", "s", &cmd_ATS);
}

void loop() {
//BT Auth
  int Push_button_state = digitalRead(PushButton);
  if ( Push_button_state == LOW )
  {
    digitalWrite(PIN_LED_R, HIGH);//Red LED OFF
    digitalWrite(PIN_LED_G, HIGH);//Gree LED OFF
    digitalWrite(PIN_LED_B, LOW);//Blue LED ON
  }
  if (confirmRequestPending)
  {
    if ( Push_button_state == LOW )
    {
      ESP_BT.confirmReply(true);
    } else {
      ESP_BT.confirmReply(false);
    }
  }


//Call Timeout Handler
  TimeOut::handler();

	//Bluetooth message received and need to be transfered via Lora
  if (ESP_BT.available() > 0) {
    // read the incoming string:
    incomingString = ESP_BT.readStringUntil('\n');
    //ATS+ command
    if(incomingString.startsWith("ATS+")) {
      //print the command
      char line[128];
      incomingString.toCharArray(line, incomingString.length());
      line[incomingString.length()] = '\0';
      char response[MyCommandParser::MAX_RESPONSE_SIZE];
      parser.processCommand(line, response);
      ESP_BT.println(response);
      return;
    }

    //Before processing Lora remove '\r' for just better terminal output
    incomingString.replace("\r","");
    // prints the received data
    Serial.println("(System [Msg]): MsgID ["+String(msgCount) +"] "+"Msg: "+String(incomingString));
    //Sent to the receiver via Lora
    sendMessage(incomingString);
    //Check the ACK after 3.5s, Resend the meessage if the ACK did not received
    ACKTimeout.timeOut(3500,callbackACKTimeout);

  }

  //Entering Receiving Mode by def
  digitalWrite(ant_switch, LOW); //Receiver mode
  //Led for receiving mode
  digitalWrite(PIN_LED_R, HIGH); //Red LED OFF
  digitalWrite(PIN_LED_G, LOW);  //Green LED ON
  digitalWrite(PIN_LED_B, HIGH); //Blue LED OFF

  // parse for a packet, and call onReceive with the result
  onReceive(LoRa.parsePacket());
}

void onReceive(int packetSize) {
  if (packetSize == 0) return;          // if there's no packet, return

  // read packet header bytes:
  byte recipient = LoRa.read();          // recipient address
  byte sender = LoRa.read();            // sender address
  byte incomingMsgId = LoRa.read();     // incoming msg ID
  byte msgType = LoRa.read();           // incoming msg Type
  byte incomingLength = LoRa.read();    // incoming msg length

  String incoming = "";

  while (LoRa.available()) {
    incoming += (char)LoRa.read();
  }

  if (incomingLength != incoming.length()) {   // check length for error
    Serial.println("error: message length does not match length");
    return;                             // skip rest of function
  }

  // if the recipient isn't this device or broadcast,
  if (recipient != localAddress && recipient != 0xFF) {
    Serial.println("This message is not for me.");
    return;                             // skip rest of function
  }

  if(msgType == (byte)ack && incomingMsgId == lastActID) {
    return; //second ACK, just ignore it
  }

  if(msgType == (byte)ack) {
    lastActID = incomingMsgId;
    Serial.println("(System [ACK]): MsgID ["+String(incomingMsgId)+"] reached ["+String(sender, HEX)+"]");
    ESP_BT.println("(System [ACK]): MsgID ["+String(incomingMsgId)+"] reached ["+String(sender, HEX)+"]");
    //Print Quality Info if necessory
    if(LoraQualityInfo == true) {
      Serial.println("(System [INF]): RSSI ["+String(LoRa.packetRssi())+"] SNR ["+String(LoRa.packetSnr())+"]");
      ESP_BT.println("(System [INF]): RSSI ["+String(LoRa.packetRssi())+"] SNR ["+String(LoRa.packetSnr())+"]");
    }
    //The message sent successfully, do not need to wait for the timeout
    if((msgCount-1) == lastActID) {
      ACKTimeout.cancel();
      retryMsg = 0;
    }
    return;                             //Stop at here for avoiding the ACK loop
  }

  //For remote Notification
  if(msgType == (byte)buzzer) {
    Serial.println("(Notification ["+String(sender, HEX)+"]): Sent you a [Buzzer]");
    ESP_BT.println("(Notification ["+String(sender, HEX)+"]): Sent you a [Buzzer]");
    //User Notification by using Buzzer and Virbation
    localNotification(buzzer, NTFRemoteRingTime);
    //return;
  }
  if(msgType == (byte)vibration) {
    Serial.println("(Notification ["+String(sender, HEX)+"]): Sent you a [Vibrator]");
    ESP_BT.println("(Notification ["+String(sender, HEX)+"]): Sent you a [Vibrator]");
    //User Notification by using Buzzer and Virbation
    localNotification(vibration, NTFRemoteRingTime);
    //return;
  }
  if(msgType == (byte)buzzer_vibration) {
    Serial.println("(Notification ["+String(sender, HEX)+"]): Sent you a [Buzzer+Vibrator]");
    ESP_BT.println("(Notification ["+String(sender, HEX)+"]): Sent you a [Buzzer+Vibrator]");
    //User Notification by using Buzzer and Virbation
    localNotification(buzzer_vibration, NTFRemoteRingTime);
    //return;
  }


//Replay ACK 2 times
  SendACK(incomingMsgId);
  delay(10);
  SendACK(incomingMsgId);

//Process the message
  if(msgType == (byte)msg) {
    Serial.println("(Message ["+String(sender, HEX)+"]): "+incoming);
    ESP_BT.println("(Message ["+String(sender, HEX)+"]): "+incoming);
    //Print Quality Info if necessory
    if(LoraQualityInfo == true) {
      Serial.println("(System [INF]): RSSI ["+String(LoRa.packetRssi())+"] SNR ["+String(LoRa.packetSnr())+"]");
      Serial.println("(System [INF]): Sender ["+String(sender, HEX)+"] Recipient ["+String(recipient, HEX)+"]");
      Serial.println("(System [INF]): msgID ["+String(incomingMsgId)+"] Length ["+String(incomingLength)+"]");
      ESP_BT.println("(System [INF]): RSSI ["+String(LoRa.packetRssi())+"] SNR ["+String(LoRa.packetSnr())+"]");
      ESP_BT.println("(System [INF]): Sender ["+String(sender, HEX)+"] Recipient ["+String(recipient, HEX)+"]");
      ESP_BT.println("(System [INF]): msgID ["+String(incomingMsgId)+"] Length ["+String(incomingLength)+"]");
    }
  }

  //User Notification by using Buzzer and Virbation
  localNotification(LocalNotificationConfig, NTFLocalRingTime);

  return;
}

void sendMessage(String outgoing) {
  //reset msgCount if it reachs 254
  if(msgCount >= 255) {
    msgCount = 0;
  }

  //Send LoRa Msg packet to receiver
  digitalWrite(ant_switch, HIGH); //Transmitter mode
  //Led for Transmitting mode
  digitalWrite(PIN_LED_R, LOW);  //Red LED ON
  digitalWrite(PIN_LED_G, HIGH); //Green LED OFF
  digitalWrite(PIN_LED_B, HIGH); //Blue LED OFF
  delay(10);

  LoRa.beginPacket();                   // start packet
  LoRa.write(destination);              // add destination address
  LoRa.write(localAddress);             // add sender address
  LoRa.write(msgCount);                 // add message ID
  LoRa.write((byte)msg);                 // add message Type
  LoRa.write(outgoing.length());        // add payload length
  LoRa.print(outgoing);                 // add payload
  LoRa.endPacket();                     // finish packet and send it
  msgCount++;                           // increment message ID

  //Entering Receiving Mode by def
  digitalWrite(ant_switch, LOW); //Receiver mode
  //Led for receiving mode
  digitalWrite(PIN_LED_R, HIGH); //Red LED OFF
  digitalWrite(PIN_LED_G, LOW);  //Green LED ON
  digitalWrite(PIN_LED_B, HIGH); //Blue LED OFF
}

void SendACK(byte msgID_Reveived)
{
  //Send LoRa ACK packet to receiver
  digitalWrite(ant_switch, HIGH); //Transmitter mode
  //Led for Transmitting mode
  digitalWrite(PIN_LED_R, LOW);  //Red LED ON
  digitalWrite(PIN_LED_G, HIGH); //Green LED OFF
  digitalWrite(PIN_LED_B, HIGH); //Blue LED OFF
  delay(10);

  outgoing =""; //Empty msg for ACK
  LoRa.beginPacket();                   // start packet
  LoRa.write(destination);              // add destination address
  LoRa.write(localAddress);             // add sender address
  LoRa.write(msgID_Reveived);           // add message ID which just received
  LoRa.write((byte)ack);                 //add message Type
  LoRa.write(outgoing.length());        // add payload length
  LoRa.print(outgoing);                 // add payload
  LoRa.endPacket();                     // finish packet and send it

  //Entering Receiving Mode by def
  digitalWrite(ant_switch, LOW); //Receiver mode
  //Led for receiving mode
  digitalWrite(PIN_LED_R, HIGH); //Red LED OFF
  digitalWrite(PIN_LED_G, LOW);  //Green LED ON
  digitalWrite(PIN_LED_B, HIGH); //Blue LED OFF
}

void callbackACKTimeout() {
  int retryDelay =  (int)random(3000, 10000);
  if((msgCount-1) != lastActID) {
    if(retryMsg != 1) {
      Serial.println("(System [INF]): MsgID ["+String(msgCount-1)+"] timeout!");
      ESP_BT.println("(System [INF]): MsgID ["+String(msgCount-1)+"] timeout!");
    }

    //resend the msg in retryDelay ms
    if(retryMsg == 0) {
      retryMsg = 1; //Resent the msg when this callback function be called next time
      // prints the received data
      incomingString.replace('\r',' ');  //Just make the terminal out looks better
      Serial.println("(System [INF]): MsgID ["+String(msgCount) +"] Retry in ["+String(int((retryDelay/1000)+0.5))+"s] "+"Msg: "+String(incomingString));
      ESP_BT.println("(System [INF]): MsgID ["+String(msgCount) +"] Retry in ["+String(int((retryDelay/1000)+0.5))+"s] "+"Msg: "+String(incomingString));

      //Resend the Msg in RetryDelay ms
      ACKTimeout.timeOut(retryDelay,callbackACKTimeout);
      return;
    }

    if(retryMsg == 1) { //resend the msg
      retryMsg = 2; //reset the retryMsg flag
      //resend the msg
      //Sent to the receiver via Lora
      sendMessage(incomingString);
      //Check the ACK after 5s, Resend the meessage if the ACK did not received
      ACKTimeout.timeOut(3500,callbackACKTimeout);
      return;
    }

    if(retryMsg == 2) {
      retryMsg = 0;
      return;
    }
  }
}

void cmd_ATS(MyCommandParser::Argument *args, char *response) {
  String Argument(args[0].asString);

  if (Argument == "B") { // Send Lora msgType 1 Buzzer to remote
    sprintf(response, "(System [NTF]): MsgID [%s] NTF [Buzzer]",String(msgCount));
    //Sent to the receiver via Lora
    sendRemoteNotification(buzzer);
    //Check the ACK after 3.5s, Resend the meessage if the ACK did not received
    ACKTimeout.timeOut(3500,callbackACKTimeout);
  }
  if (Argument == "V") { //Send Lora msgType 2 Vibration to remote
    sprintf(response, "(System [NTF]): MsgID [%s] NTF [Vibration]",String(msgCount));
    //Sent to the receiver via Lora
    sendRemoteNotification(vibration);
    //Check the ACK after 3.5s, Resend the meessage if the ACK did not received
    ACKTimeout.timeOut(3500,callbackACKTimeout);
  }
  if (Argument == "BV") { //Send Lora msgType 3 Buzzer + Vibration to remote
    sprintf(response, "(System [NTF]): MsgID [%s] NTF [Buzzer+Vibration]",String(msgCount));
    //Sent to the receiver via Lora
    sendRemoteNotification(buzzer_vibration);
    //Check the ACK after 3.5s, Resend the meessage if the ACK did not received
    ACKTimeout.timeOut(3500,callbackACKTimeout);
  }
  if (Argument == "INF") { //Trun ON/OFF Local LoraQualityInfo
    LoraQualityInfo = !LoraQualityInfo;

    if(LoraQualityInfo == true) {
      strlcpy(response, "(System [CMD]): LoraQualityInfo [ON]", MyCommandParser::MAX_RESPONSE_SIZE);
    } else {
      strlcpy(response, "(System [CMD]): LoraQualityInfo [OFF]", MyCommandParser::MAX_RESPONSE_SIZE);
    }
  }
  if (Argument == "NTFB") { //Configure Local Notification to Buzzer
    LocalNotificationConfig = 1;
    strlcpy(response, "(System [CMD]): Local NTF [Buzzer]", MyCommandParser::MAX_RESPONSE_SIZE);
  }
  if (Argument == "NTFV") { //Configure Local Notification to Virbation
    LocalNotificationConfig = 2;
    strlcpy(response, "(System [CMD]): Local NTF [Vibration]", MyCommandParser::MAX_RESPONSE_SIZE);
  }
  if (Argument == "NTFBV") { //Configure Local Notification to Buzzer+Virbation
    LocalNotificationConfig = 3;
    strlcpy(response, "(System [CMD]): Local NTF [Buzzer+Vibration]", MyCommandParser::MAX_RESPONSE_SIZE);
  }
  if (Argument == "NTFM") { //Configure Local Notification to Mute
    LocalNotificationConfig = 0;
    strlcpy(response, "(System [CMD]): Local NTF [Mute]", MyCommandParser::MAX_RESPONSE_SIZE);
  }
  if (Argument == "NTFT") { //Configure Local Notification Testing
    //Testing the current Local Notification Configuration
    localNotification(LocalNotificationConfig,1000); //Testing for 1s
    strlcpy(response, "(System [CMD]): Local NTF [Testing]", MyCommandParser::MAX_RESPONSE_SIZE);
  }
}

void sendRemoteNotification(byte msgType) //msgType could be 1 for buzzer, 2 for Vibration,3 for Buzzer + Virbation
{
  //reset msgCount if it reachs 254
  if(msgCount >= 255) {
    msgCount = 0;
  }

  //Send LoRa ACK packet to receiver
  digitalWrite(ant_switch, HIGH); //Transmitter mode
  //Led for Transmitting mode
  digitalWrite(PIN_LED_R, LOW);  //Red LED ON
  digitalWrite(PIN_LED_G, HIGH); //Green LED OFF
  digitalWrite(PIN_LED_B, HIGH); //Blue LED OFF
  delay(10);

  outgoing =""; //Empty msg for ACK
  LoRa.beginPacket();                   // start packet
  LoRa.write(destination);              // add destination address
  LoRa.write(localAddress);             // add sender address
  LoRa.write(msgCount);           // add message ID which just received
  LoRa.write(msgType);                 //add message Type
  LoRa.write(outgoing.length());        // add payload length
  LoRa.print(outgoing);                 // add payload
  LoRa.endPacket();                     // finish packet and send it
  msgCount++;                           // increment message ID

  //Entering Receiving Mode by def
  digitalWrite(ant_switch, LOW); //Receiver mode
  //Led for receiving mode
  digitalWrite(PIN_LED_R, HIGH); //Red LED OFF
  digitalWrite(PIN_LED_G, LOW);  //Green LED ON
  digitalWrite(PIN_LED_B, HIGH); //Blue LED OFF
}

void localNotification(byte LocalNTFConfig, int RingTime)
{
  switch(LocalNTFConfig) {
  case 0: //mute
    break;
  case 1: // Local Notification Buzzer
    digitalWrite(PIN_Buzzer, HIGH); //Turn on Buzzer
    NTFTimeout.timeOut(RingTime, callbackNTFTimeout);
    break;
  case 2: //Local Notification Virbation
    digitalWrite(PIN_Vibration, HIGH); //Turn on Virbation
    NTFTimeout.timeOut(RingTime, callbackNTFTimeout);
    break;
  case 3: //Local Notification Buzzer+Virbation
    digitalWrite(PIN_Buzzer, HIGH); //Turn on Buzzer
    digitalWrite(PIN_Vibration, HIGH); //Turn on Virbation
    NTFTimeout.timeOut(RingTime, callbackNTFTimeout);
    break;
  default:
    break;
  }
}

void callbackNTFTimeout() {
  //Turn off Buzzer and Virbration
  digitalWrite(PIN_Buzzer, LOW);
  digitalWrite(PIN_Vibration, LOW);
}

void BTConfirmRequestCallback(uint32_t numVal)
{
  confirmRequestPending = true;
  Serial.println(numVal);
}

void BTAuthCompleteCallback(boolean success)
{
  confirmRequestPending = false;
  if (success)
  {
    Serial.println("Pairing success!!");
    localNotification(LocalNotificationConfig, NTFLocalRingTime);
  }
  else
  {
    Serial.println("Pairing failed, rejected by user!!");
  }
}
